#ifndef RW_LOCK
#define RW_LOCK

#define __STDC_LIMIT_MACROS

#define MARKER_DIST 2
#define HMCS_THRESHOLD 50
#define TREE_THRESHOLD 50
#define READER_THRESHOLD 4000
#define WRITER_ROOT_THRESHOLD 20
#define DISTRIBUTION 1000
#define MEMORY_LEVELS 2

#include "fompi.h"
#include <inttypes.h>
#include <stdint.h>

using namespace std;

//#define PRINT 0
//#define __DEBUG

extern int marker_dist, hmcs_threshold, tree_threshold, root_threshold, reader_threshold, writer_rate;

enum {COHORT_START = 0, CHANGE_MODE = INT64_MAX - 1, ACQUIRE_PARENT = INT64_MAX - 2, WAIT = INT64_MAX, ACQUIRE_GLOBAL = INT64_MAX - 1};
enum {WRITE = INT64_MAX / 2, START_READ = 0};
enum {INC = 0, DEC = 1};
enum {NEXT = 0, STATUS = 1, TAIL = 2, PARENT = 3, THRESHOLD = 4, WRITER_THRESHOLD = 5};

extern int64_t rank_world, rank_node, rank_hNode;

void check_and_wait_for_write(foMPI_Win win_marker);
void change_markers(int64_t result, foMPI_Win win_marker);
void check_markers(foMPI_Win win_marker);
void reset_one_marker(int rank_of_marker, foMPI_Win win_marker);
void reset_markers(foMPI_Win win_marker);

template<int depth> 
struct WLock {
	
	/** Acquiring of lock on level N
	*/
	inline void acquire(int64_t *hNode, int64_t *qNode, foMPI_Win win_hNode, foMPI_Win win_qNode, foMPI_Win win_tail, foMPI_Win win_marker) {
		int64_t current_status, wait = WAIT, null = -1, cohort_start = COHORT_START;
		int64_t predecessor;
		WLock<depth-1> w_lock;
		
		#ifdef PRINT
		printf("%" PRId64 " before acquire level N\n", rank_world);
		#endif
		/* Prepare node for use */
		qNode[STATUS] = WAIT;
		qNode[NEXT] = -1;
		
		/* Put yourself at end of queue */
		foMPI_Fetch_and_op(&rank_node, &predecessor, MPI_INT64_T, rank_hNode, TAIL, foMPI_REPLACE, win_hNode);
		foMPI_Win_flush(rank_hNode, win_hNode);
		
		#ifdef PRINT
		printf("%" PRId64 " level N after fetch and op, predecessor: %" PRId64 "\n", rank_world, predecessor);
		#endif
		
		/* Check if there is a predecessor and if yes put yourself as next and wait on signal */
		if(predecessor != -1){
			foMPI_Accumulate(&rank_node, 1, MPI_INT64_T, predecessor, NEXT, 1, MPI_INT64_T, foMPI_REPLACE, win_qNode);
			foMPI_Win_flush(predecessor, win_qNode);
			#ifdef PRINT
			printf("%" PRId64 " level N before sync\n", rank_world);
			#endif
			do {
				foMPI_Win_sync(win_qNode);
			} while ((current_status = qNode[STATUS]) == WAIT);
			
			/* Maybe the predecessor reached the hierarchical threshold */
			if(current_status != ACQUIRE_PARENT){
				#ifdef PRINT
				printf("%" PRId64 " level N acquired\n", rank_world);
				#endif
				return;
			}
		}
		#ifdef PRINT
		printf("%" PRId64 " level N before acquire parent\n", rank_world);
		#endif
		
		/* Acquire lock one level above */
		qNode[STATUS] = COHORT_START;
		w_lock.acquire(win_hNode, win_qNode, win_tail, win_marker);
	}

	/** Releasing of lock at level N
	*/
	inline void release(int64_t *hNode, int64_t *qNode, foMPI_Win win_hNode, foMPI_Win win_qNode, foMPI_Win win_tail, foMPI_Win win_marker) {
		int64_t current_status, successor, null_rank = -1, current_tail, acquire_parent = ACQUIRE_PARENT;
		int64_t threshold;
		WLock<depth-1> w_lock;
		
		#ifdef PRINT
		printf("%" PRId64 " before release level N\n", rank_world);
		#endif
		
		foMPI_Win_sync(win_qNode);
		successor = qNode[NEXT];
		
		#ifdef PRINT
		printf("%" PRId64 " successor: %" PRId64 ", status: %" PRId64 "\n", rank_world, successor, qNode[STATUS]);
		#endif
		
		if(successor != -1 && qNode[STATUS] < hNode[THRESHOLD]){
			/* We have successor thus pass it to the successor if no threshold is reached */
			current_status = qNode[STATUS] + 1;
			#ifdef PRINT
			printf("%" PRId64 " lock passed, successor: %" PRId64 "\n", rank_world, successor);
			#endif
			foMPI_Accumulate(&current_status, 1, MPI_INT64_T, successor, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, win_qNode);
			foMPI_Win_flush(successor, win_qNode);
			return;
		}
		/* Release parent level */
		w_lock.release(win_hNode, win_qNode, win_tail, win_marker);
		/* Check for successor, notify them to acquire parent otherwise reset the tail of the queue */
		#ifdef PRINT
		printf("%" PRId64 " release helper, successor: %" PRId64 "\n", rank_world, successor);
		#endif
		if(successor == -1){
			foMPI_Compare_and_swap(&null_rank, &rank_node, &current_tail, MPI_INT64_T, rank_hNode, TAIL, win_hNode);
			foMPI_Win_flush(rank_hNode, win_hNode);
			#ifdef PRINT
			printf("%" PRId64 " release after CAS, current_tail: %" PRId64 "\n", rank_world, current_tail);
			#endif
			if(current_tail == rank_node) {
				#ifdef PRINT
				printf("%" PRId64 " lock level N released to nobody\n", rank_world);
				#endif
				return;
			}
			#ifdef PRINT
			printf("%" PRId64 " before sync\n", rank_world);
			#endif
			do {
				foMPI_Win_sync(win_qNode);
			} while ((successor = qNode[NEXT]) == -1);
		}
		foMPI_Accumulate(&acquire_parent, 1, MPI_INT64_T, successor, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, win_qNode);
		foMPI_Win_flush(successor, win_qNode);
		#ifdef PRINT
		printf("%" PRId64 " after release level N, successor: %" PRId64 ", val: %" PRIX64 "\n", rank_world, successor, acquire_parent);
		#endif
	}
};

template<> 
struct WLock<1>{
	/** Acquiring of lock at top level 
	*/
	inline void acquire(foMPI_Win win_hNode, foMPI_Win win_qNode, foMPI_Win win_tail, foMPI_Win win_marker) {
		int64_t node[2];
		#ifdef PRINT
		printf("%" PRId64 " level 1 acquire\n", rank_world);
		#endif
		int64_t predecessor, current_status;
		
		/* Prepare node to use */
		node[STATUS] = WAIT;
		node[NEXT] = -1;
		foMPI_Accumulate(node, 2, MPI_INT64_T, rank_hNode, NEXT, 2, MPI_INT64_T, foMPI_REPLACE, win_hNode);
		foMPI_Win_flush(rank_hNode, win_hNode);
		
		/* Check for a predecessor */
		foMPI_Fetch_and_op(&rank_hNode, &predecessor, MPI_INT64_T, 0, TAIL, foMPI_REPLACE, win_tail);
		foMPI_Win_flush(0, win_tail);
		#ifdef PRINT
		printf("%" PRId64 " level 1 after fetch and op, predecessor: %" PRId64 ", rank_hNode: %" PRId64 "\n", rank_world, predecessor, rank_hNode);
		#endif
		if(predecessor == -1){	
			/* Lock acquired if mode on WRITE otherwise change it */
			check_and_wait_for_write(win_marker);
			current_status = COHORT_START;
			foMPI_Accumulate(&current_status, 1, MPI_INT64_T, rank_hNode, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, win_hNode);
			foMPI_Win_flush(rank_hNode, win_hNode);
		} else {
			/* Notify predecessor that you are next */
			foMPI_Accumulate(&rank_hNode, 1, MPI_INT64_T, predecessor, NEXT, 1, MPI_INT64_T, foMPI_REPLACE, win_hNode);
			foMPI_Win_flush(predecessor, win_hNode);
			#ifdef PRINT
			printf("%" PRId64 " level 1 sync\n", rank_world);
			#endif
			do {
				foMPI_Get_accumulate(0, 0, MPI_INT64_T, &current_status, 1, MPI_INT64_T, rank_hNode, STATUS, 1, MPI_INT64_T, foMPI_NO_OP, win_hNode);
				foMPI_Win_flush(rank_hNode, win_hNode);
			} while (current_status == WAIT);
			/* Check if predecessor changed mode to READ and if yes, change it back */
			if(current_status == CHANGE_MODE){
				check_and_wait_for_write(win_marker);
				current_status = COHORT_START;
				foMPI_Accumulate(&current_status, 1, MPI_INT64_T, rank_hNode, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, win_hNode);
				foMPI_Win_flush(rank_hNode, win_hNode);
			}
		}
		/* Lock acquired */
		#ifdef PRINT
		printf("%" PRId64 " level 1 acquired\n", rank_world);
		#endif
	}

	/** Releasing of lock at top level
	*/
	inline void release(foMPI_Win win_hNode, foMPI_Win win_qNode, foMPI_Win win_tail, foMPI_Win win_marker) {
		int64_t null_rank = -1;
		int64_t current_tail, current_status;
		int64_t successor;
		bool is_reset = false;
		
		/* Check if threshold is reached and change mode if necessary */
		foMPI_Get_accumulate(0, 0, MPI_INT64_T, &current_status, 1, MPI_INT64_T, rank_hNode, STATUS, 1, MPI_INT64_T, foMPI_NO_OP, win_hNode);
		foMPI_Win_flush(rank_hNode, win_hNode);
		current_status++;
		if(current_status == root_threshold){
			#ifdef PRINT
			printf("%" PRId64 " writer threshold reached!\n", rank_world);
			#endif
			reset_markers(win_marker);
			is_reset = true;
			current_status = CHANGE_MODE;
		}
		
		/* Check for successor, pass the value to them otherwise reset the tail of the queue */
		foMPI_Get_accumulate(0, 0, MPI_INT64_T, &successor, 1, MPI_INT64_T, rank_hNode, NEXT, 1, MPI_INT64_T, foMPI_NO_OP, win_hNode);
		foMPI_Win_flush(rank_hNode, win_hNode);
		#ifdef PRINT
		printf("%" PRId64 " level 1 release, successor: %" PRId64 ", rank_hNode: %" PRId64 "\n", rank_world, successor, rank_hNode);
		#endif
		if(successor == -1){
			if(!is_reset){
				reset_markers(win_marker);
				current_status = CHANGE_MODE;
			}
			foMPI_Compare_and_swap(&null_rank, &rank_hNode, &current_tail, MPI_INT64_T, 0, TAIL, win_tail);
			foMPI_Win_flush(0, win_tail);
			
			#ifdef PRINT
			printf("%" PRId64 " release after CAS, current_tail: %" PRId64 "\n", rank_world, current_tail);
			#endif
			if(current_tail == rank_hNode) {
				#ifdef PRINT
				printf("%" PRId64 " lock level 1 released to nobody\n", rank_world);
				#endif
				return;
			}
			#ifdef PRINT
			printf("%" PRId64 " before sync\n", rank_world);
			#endif
			do {
				foMPI_Get_accumulate(0, 0, MPI_INT64_T, &successor, 1, MPI_INT64_T, rank_hNode, NEXT, 1, MPI_INT64_T, foMPI_NO_OP, win_hNode);
				foMPI_Win_flush(rank_hNode, win_hNode);
			} while (successor == -1);
		}
		foMPI_Accumulate(&current_status, 1, MPI_INT64_T, successor, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, win_hNode);
		foMPI_Win_flush(successor, win_hNode);
		#ifdef PRINT
		printf("%" PRId64 " level 1 released, successor: %" PRId64 ", val: %" PRIX64 "\n", rank_world, successor, current_status);
		#endif
		return;
	}
};

class MCSLock {
	int64_t *lmem;
	int64_t rank;
	foMPI_Win win;
	
	public:
	void initialize();
	void acquire();
	void release();
	void finalize();
};

class HMCSLock {
	int64_t *local_mem, *global_mem;
	int64_t local_rank, global_rank, local_tail_rank;
	foMPI_Win local_win, global_win;
	
	public:
	void initialize();
	void acquire();
	void release();
	void finalize();
};

class RWLock {
	int node_size, temp_rank;
	
	int64_t *hNode, *qNode, tailNode[6];
	foMPI_Win win_hNode, win_qNode, win_tail;
	MPI_Comm comm_world, comm_node;
	
	int64_t one = 1;

	int rank_marker;
	int64_t marker[2];
	foMPI_Win win_marker;
	
	WLock<MEMORY_LEVELS> w_lock;
	
	public:
	void initialize();
	void reader_acquire();
	void reader_release();
	void writer_acquire();
	void writer_release();
	void finalize();
};

#endif