#include "Locks.h"
#include "libtopodisc.h"

using namespace std;

int64_t rank_world, rank_node, rank_hNode;
int world_size;

/** Initialization of the hierarchical tree with a queue and the markers for a RW-lock 
*/
void RWLock::initialize()
{
	int64_t node[6];
	
	comm_world = MPI_COMM_WORLD;
    MPI_Comm_rank(comm_world, &temp_rank);
	rank_world = (int64_t) temp_rank;
	MPI_Comm_size(comm_world, &world_size);
	
	/* Set up of communicator and window for HNode on one node */
	TPD_Get_hier_levels(&comm_node, 1, comm_world);
    foMPI_Win_allocate(6 * sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm_world, &hNode, &win_hNode);
	
	hNode[PARENT] = 0;
	hNode[NEXT] = -1;
	hNode[STATUS] = -1;
	hNode[TAIL] = -1;
	hNode[THRESHOLD] = tree_threshold;
	hNode[WRITER_THRESHOLD] = -1;
	
	/* Set up QNodes */
    foMPI_Win_allocate(2 * sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm_node, &qNode, &win_qNode);
	qNode[NEXT] = -1;
	qNode[STATUS] = -1;
	
	/* Define marker nodes */
	MPI_Comm_size(comm_node, &node_size);
	MPI_Comm_rank(comm_node, &temp_rank);
	rank_node = (int64_t) temp_rank;
	rank_marker = (int) (rank_world - (rank_world % ((int64_t) marker_dist)));
	
	/* Set up window for the markers */
	foMPI_Win_allocate(2 * sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm_world, marker, &win_marker);	
	node[INC] = 0;
	node[DEC] = 0;
	foMPI_Accumulate(node, 2, MPI_INT64_T, rank_world, 0, 2, MPI_INT64_T, foMPI_REPLACE, win_marker);
	foMPI_Win_flush(rank_world, win_marker);
	
	/* Set up window for root of hierarchy tree */
	foMPI_Win_allocate(6 * sizeof(int64_t), sizeof(int64_t), MPI_INFO_NULL, comm_world, tailNode, &win_tail);	
	node[PARENT] = -1;
	node[NEXT] = -1;
	node[STATUS] = -1;
	node[TAIL] = -1;
	node[THRESHOLD] = (int64_t) tree_threshold;
	node[WRITER_THRESHOLD] = (int64_t) root_threshold;
	foMPI_Accumulate(node, 6, MPI_INT64_T, rank_world, 0, 6, MPI_INT64_T, foMPI_REPLACE, win_tail);
	foMPI_Win_flush(rank_world, win_tail);
	
	rank_hNode = rank_world - (rank_world % (int64_t) node_size);
	
	if(rank_world == (int64_t) 0){
		printf("WORLD SIZE: %i, NODE SIZE: %i, WRITER_ROOT_THRESHOLD: %i, TREE_THRESHOLD: %i, READER_THRESHOLD: %i, RATE: %i, MARKER_DIST: %i\n----------------------------------------------------\n", world_size, node_size, root_threshold, tree_threshold, reader_threshold, writer_rate, marker_dist);
	}
	
	#ifdef PRINT
	printf("%" PRId64 " INIT, rank_node: %" PRId64 ", rank_marker: %i, rank_hNode: %" PRId64 ", node_size: %i\n", rank_world, rank_node, rank_marker, rank_hNode, node_size);
	#endif

    foMPI_Win_fence(0, win_hNode);
    foMPI_Win_fence(0, win_qNode);
    foMPI_Win_fence(0, win_marker);
    foMPI_Win_fence(0, win_tail); 
}

/** Free everything up
*/
void RWLock::finalize()
{
	foMPI_Win_fence(0, win_hNode);
    foMPI_Win_fence(0, win_qNode);
    foMPI_Win_fence(0, win_marker);
    foMPI_Win_fence(0, win_tail); 
	
	foMPI_Win_free(&win_hNode);
	foMPI_Win_free(&win_qNode);
	foMPI_Win_free(&win_marker);
	foMPI_Win_free(&win_tail);
	
	MPI_Comm_free(&comm_node);
}

void RWLock::writer_acquire() {
	w_lock.acquire(hNode, qNode, win_hNode, win_qNode, win_tail, win_marker);
}

void RWLock::writer_release() {
	w_lock.release(hNode, qNode, win_hNode, win_qNode, win_tail, win_marker);
}

/** Acquire a reader lock
*/
void RWLock::reader_acquire() {
	int64_t count, current_tail, minus = -1;
	bool barrier = false;
	start:
	#ifdef PRINT
	printf("%" PRId64 " reader acquire\n", rank_world);
	#endif
	/* If barrier raised wait until marker is reset */
	if(barrier){
		do {
			foMPI_Get_accumulate(0, 0, MPI_INT64_T, &count, 1, MPI_INT64_T, rank_marker, INC, 1, MPI_INT64_T, foMPI_NO_OP, win_marker);
			foMPI_Win_flush(rank_marker, win_marker);
		} while (count > reader_threshold);
		#ifdef PRINT
		printf("%" PRId64 " barrier released, count: %" PRIX64 "\n", rank_world, count);
		#endif
	}
	/* Add one to the counter and fetch it */
	foMPI_Fetch_and_op(&one, &count, MPI_INT64_T, rank_marker, INC, foMPI_SUM, win_marker);
	foMPI_Win_flush(rank_marker, win_marker);
	if(count >= reader_threshold){
		barrier = true;
		#ifdef PRINT
		printf("%" PRId64 " reader over threshold, marker: %i, count: %" PRIX64 "\n", rank_world, rank_marker, count);
		#endif
		/* Decrement counter and check if you were the first over the threshold */
        if(count == reader_threshold){
			foMPI_Get_accumulate(0, 0, MPI_INT64_T, &current_tail, 1, MPI_INT64_T, 0, TAIL, 1, MPI_INT64_T, foMPI_NO_OP, win_tail);
			foMPI_Win_flush(0, win_tail);
			#ifdef PRINT
			printf("%" PRId64 " threshold reached, tail: %" PRId64 ", count: %" PRIX64 "\n", rank_world, current_tail, count);
			#endif
			/* Check if there is a writer to pass the lock otherwise keep it */
			if(current_tail == -1){
				reset_one_marker(rank_marker, win_marker);
				barrier = false;
			}
		}
		foMPI_Accumulate(&minus, 1, MPI_INT64_T, rank_marker, INC, 1, MPI_INT64_T, foMPI_SUM, win_marker);
		foMPI_Win_flush(rank_marker, win_marker);
		goto start;
	}
	#ifdef PRINT
	printf("%" PRId64 " reader lock acquired, count: %" PRIX64 "\n", rank_world, count);
	#endif
}

/** Release of a reader-lock
*/
void RWLock::reader_release(){
	/* Decrement counter */
	foMPI_Accumulate(&one, 1, MPI_INT64_T, rank_marker, DEC, 1, MPI_INT64_T, foMPI_SUM, win_marker);	
	foMPI_Win_flush(rank_marker, win_marker);
	#ifdef PRINT
	printf("%" PRId64 " reader release\n", rank_world);
	#endif
}

/** Checks the marker and if it is not on WRITE, it will change it to WRITE
*/
void check_and_wait_for_write(foMPI_Win win_marker){
	/* If mode is READ change it to WRITE if no reader is active */
	check_markers(win_marker);
	change_markers(WRITE, win_marker);
	check_markers(win_marker);
}

/** Change all markers in the system to the given value
*/
void change_markers(int64_t result, foMPI_Win win_marker){
	int i;
	#ifdef PRINT
	printf("%" PRId64 " before change markers, result: %" PRIX64 "\n", rank_world, result);
	#endif
	for (i = 0; i < world_size; i += marker_dist){
		foMPI_Accumulate(&result, 1, MPI_INT64_T, i, INC, 1, MPI_INT64_T, foMPI_SUM, win_marker);
	}
	foMPI_Win_flush_all(win_marker);
}

/** Check all markers if somebody is still holding a reader-lock and wait on them
*/
void check_markers(foMPI_Win win_marker){
	int i;
	int64_t counters[2];
	#ifdef PRINT
	printf("%" PRId64 " before check markers, world_size: %i\n", rank_world, world_size);
	#endif
	for (i = 0; i < world_size; i += marker_dist){
		do {
			foMPI_Get_accumulate(0, 0, MPI_INT64_T, counters, 2, MPI_INT64_T, i, INC, 2, MPI_INT64_T, foMPI_NO_OP, win_marker);
			foMPI_Win_flush(i, win_marker);
			#ifdef PRINT
			printf("%" PRId64 " check marker: %i, inc: %" PRIX64 ", dec: %" PRIX64 "\n", rank_world, i, counters[INC], counters[DEC]);
			#endif
			if(counters[INC] - counters[DEC] - WRITE == 0)
				break;
		} while(counters[INC] - counters[DEC] > 0);
	}
}

/** Reset one marker with given rank 
*/
void reset_one_marker(int rank_of_marker, foMPI_Win win_marker){
	int64_t counters[2];
	int64_t sub_counters[2];
	
	#ifdef PRINT
	printf("%" PRId64 " reset marker: %i\n", rank_world, rank_of_marker);
	#endif
	/* Get the current count */
	foMPI_Get_accumulate(0, 0, MPI_INT64_T, counters, 2, MPI_INT64_T, rank_of_marker, INC, 2, MPI_INT64_T, foMPI_NO_OP, win_marker);
	foMPI_Win_flush(rank_of_marker, win_marker);
	
	/* Subtract the decrement counter from both because this is the number of successful acquirers */
	sub_counters[INC] = -counters[DEC];
	sub_counters[DEC] = -counters[DEC];
	
	/* Reset WRITE if set */
	if(counters[INC] >= WRITE){
		sub_counters[INC] -= WRITE;
	}
	#ifdef PRINT
	printf("%" PRId64 " reset marker: %i, inc: %" PRIx64 ", dec: %" PRIx64 ", subinc: %" PRIX64 ", subdec: %" PRIx64 "\n", rank_world, rank_of_marker, counters[INC], counters[DEC], sub_counters[INC], sub_counters[DEC]);
	#endif
	foMPI_Accumulate(sub_counters, 2, MPI_INT64_T, rank_of_marker, INC, 2, MPI_INT64_T, foMPI_SUM, win_marker);
	foMPI_Win_flush(rank_of_marker, win_marker);
}

/** Reset all the markers 
*/
void reset_markers(foMPI_Win win_marker){
	int i;
	for (i = 0; i < world_size; i += marker_dist){
		reset_one_marker(i, win_marker);
	}
}
