#include "fompi.h"
#include "Locks.h"
#include "libtopodisc.h"

/* Initialization of window for a MCS-queue */
void HMCSLock::initialize()
{
    MPI_Aint winsize;
	int temp_rank, local_size;
	MPI_Comm local_comm;
	
	MPI_Comm global_comm = MPI_COMM_WORLD;
	TPD_Get_hier_levels(&local_comm, 1, global_comm);
	
    MPI_Comm_rank(global_comm, &temp_rank);
	global_rank = (int64_t) temp_rank;
	MPI_Comm_rank(local_comm, &temp_rank);
	local_rank = (int64_t) temp_rank;
	
	MPI_Comm_size(local_comm, &local_size);
	local_tail_rank = (global_rank / local_size) * local_size;
	
	#ifdef PRINT
	printf("%" PRId64 " HMCS init, local_rank: %" PRId64 ", local_tail_rank: %" PRId64 "\n", global_rank, local_rank, local_tail_rank);
	#endif
 
    winsize = 3 * sizeof(int64_t);
    foMPI_Win_allocate(winsize, sizeof(int64_t), MPI_INFO_NULL, local_comm, &local_mem, &local_win);
	foMPI_Win_allocate(winsize, sizeof(int64_t), MPI_INFO_NULL, global_comm, &global_mem, &global_win);
    local_mem[NEXT] = -1;
    local_mem[STATUS] = 0;
    local_mem[TAIL] = -1;
    global_mem[NEXT] = -1;
    global_mem[STATUS] = 0;
    global_mem[TAIL] = -1;
    foMPI_Win_fence(0, local_win);
    foMPI_Win_fence(0, global_win);
	foMPI_Win_lock_all(0, local_win);    
	foMPI_Win_lock_all(0, global_win); 	
	#ifdef PRINT
	printf("%" PRId64 " HMCS initialized\n", global_rank);
	#endif
}

void HMCSLock::acquire()
{
    int64_t predecessor, status;
	int64_t null = -1, wait = WAIT;
	
	local_mem[STATUS] = wait;
	local_mem[NEXT] = null;
	
	/* Add yourself to end of local queue */
    foMPI_Fetch_and_op(&local_rank, &predecessor, MPI_INT64_T, 0, TAIL, foMPI_REPLACE, local_win);
	foMPI_Win_flush(0, local_win);

	#ifdef PRINT
	printf("%" PRId64 " LOCAL predecessor: %" PRId64 "\n", global_rank, predecessor);
	#endif
	
    if(predecessor != -1)
    {
        /* We didn't get lock. Add us as next at predecessor. */
        foMPI_Accumulate(&local_rank, 1, MPI_INT64_T, predecessor, NEXT, 1, MPI_INT64_T, foMPI_REPLACE, local_win);
		foMPI_Win_flush(predecessor, local_win);
		
		#ifdef PRINT
		printf("%" PRId64 " LOCAL sync predecessor: %" PRId64 "\n", global_rank, predecessor);
		#endif
		
		/* Now spin on our local value "status" until we are given the lock. */
        do {
            foMPI_Win_sync(local_win);
        } while ((status = local_mem[STATUS]) == WAIT);
		if(status != ACQUIRE_GLOBAL) {
			#ifdef PRINT
			printf("%" PRId64 " LOCAL lock acquired\n", global_rank);
			#endif
			return;
		}
    }
	foMPI_Accumulate(&wait, 1, MPI_INT64_T, local_tail_rank, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, global_win);
	foMPI_Accumulate(&null, 1, MPI_INT64_T, local_tail_rank, NEXT, 1, MPI_INT64_T, foMPI_REPLACE, global_win);
	foMPI_Win_flush(local_tail_rank, global_win);
	/* Add yourself to end of global queue */
    foMPI_Fetch_and_op(&local_tail_rank, &predecessor, MPI_INT64_T, 0, TAIL, foMPI_REPLACE, global_win);
	foMPI_Win_flush(0, global_win);
	
	#ifdef PRINT
	printf("%" PRId64 " GLOBAL predecessor: %" PRId64 "\n", global_rank, predecessor);
	#endif
	
	if(predecessor != -1)
    {
        /* We didn't get lock. Add us as next at predecessor. */
        foMPI_Accumulate(&local_tail_rank, 1, MPI_INT64_T, predecessor, NEXT, 1, MPI_INT64_T, foMPI_REPLACE, global_win);
		
		#ifdef PRINT
		printf("%" PRId64 " GLOBAL sync, predecessor: %" PRId64 "\n", global_rank, predecessor);
		#endif
		
		/* Now spin on our local value "status" until we are given the lock. */
        do {
            foMPI_Get_accumulate(0, 0, MPI_INT64_T, &status, 1, MPI_INT64_T, local_tail_rank, STATUS, 1, MPI_INT64_T, foMPI_NO_OP, global_win);
			foMPI_Win_flush(local_tail_rank, global_win);
        } while (status == WAIT);
    }
	local_mem[STATUS] = COHORT_START;
	#ifdef PRINT
	printf("%" PRId64 " GLOBAL lock acquired\n", global_rank);
	#endif
}

void HMCSLock::release()
{
    int64_t nullrank = -1, zero = 0, curtail = -1;
	int64_t local_successor, global_successor, current_status, acquire_global = ACQUIRE_GLOBAL;

	foMPI_Win_sync(local_win);
	#ifdef PRINT
	printf("%" PRId64 " LOCAL release successor: %" PRId64 ", status: %" PRId64 "\n", global_rank, local_mem[NEXT], local_mem[STATUS]);
	#endif
	local_successor = local_mem[NEXT];
	
    if (local_successor != -1 && local_mem[STATUS] < HMCS_THRESHOLD){
			/* Notify the successor */
			current_status = local_mem[STATUS] + 1;
			foMPI_Accumulate(&current_status, 1, MPI_INT64_T, local_successor, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, local_win);
			foMPI_Win_flush(local_successor, local_win);
			#ifdef PRINT
			printf("%" PRId64 " LOCAL released to successor: %" PRId64 ", status: %" PRIx64 "\n", global_rank, local_successor, current_status);
			#endif
			return;
	}
	foMPI_Get_accumulate(0, 0, MPI_INT64_T, &global_successor, 1, MPI_INT64_T, local_tail_rank, NEXT, 1, MPI_INT64_T, foMPI_NO_OP, global_win);
	foMPI_Win_flush(local_tail_rank, global_win);
	#ifdef PRINT
	printf("%" PRId64 " GLOBAL release successor: %" PRId64 "\n", global_rank, global_successor);
	#endif
	if(global_successor == -1) {
		/* See if we're waiting for the next to notify us */
		foMPI_Compare_and_swap(&nullrank, &local_tail_rank, &curtail, MPI_INT64_T, 0, TAIL, global_win);
		foMPI_Win_flush(0, global_win);
		#ifdef PRINT
		printf("%" PRId64 " GLOBAL CAS curtail: %" PRId64 "\n", global_rank, curtail);
		#endif
		if(curtail != local_tail_rank) {
			#ifdef PRINT
			printf("%" PRId64 " GLOBAL sync release\n", global_rank);
			#endif
			do {
				foMPI_Get_accumulate(0, 0, MPI_INT64_T, &global_successor, 1, MPI_INT64_T, local_tail_rank, NEXT, 1, MPI_INT64_T, foMPI_NO_OP, global_win);
				foMPI_Win_flush(local_tail_rank, global_win);
			} while (global_successor == -1);
		}
	}
	if(global_successor != -1 && curtail != local_tail_rank){
		foMPI_Accumulate(&zero, 1, MPI_INT64_T, global_successor, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, global_win);
		foMPI_Win_flush(global_successor, global_win);
		#ifdef PRINT
		printf("%" PRId64 " GLOBAL released to successor: %" PRId64 "\n", global_rank, global_successor);
		#endif
	}
	#ifdef PRINT
	if(global_successor == -1){
		printf("%" PRId64 " GLOBAL released, no successor\n", global_rank);
	}
	#endif
	
	if(local_successor == -1) {
		/* See if we're waiting for the next to notify us */
		foMPI_Compare_and_swap(&nullrank, &local_rank, &curtail, MPI_INT64_T, 0, TAIL, local_win);
		foMPI_Win_flush(0, local_win);
		#ifdef PRINT
		printf("%" PRId64 " LOCAL CAS curtail: %" PRId64 "\n", global_rank, curtail);
		#endif
		if(curtail == local_rank)
		{
			#ifdef PRINT
			printf("%" PRId64 " LOCAL released, no successor \n", global_rank);
			#endif
			return;
		}
		/* Someone else has added themselves to the list. */
		#ifdef PRINT
		printf("%" PRId64 " LOCAL sync release\n", global_rank);
		#endif
		do {
            foMPI_Win_sync(local_win);
        } while ((local_successor = local_mem[NEXT]) == -1);
	}
	
	#ifdef PRINT
	printf("%" PRId64 " LOCAL notify successor acquire global: %" PRId64 "\n", global_rank, local_successor);
	#endif
	foMPI_Accumulate(&acquire_global, 1, MPI_INT64_T, local_successor, STATUS, 1, MPI_INT64_T, foMPI_REPLACE, local_win);
	foMPI_Win_flush(local_successor, local_win);
}

/* Free window for MCS-queue */
void HMCSLock::finalize()
{
	#ifdef PRINT
	printf("%" PRId64 " HMCS Finalization\n", global_rank);
	#endif
	foMPI_Win_unlock_all(local_win);
	foMPI_Win_unlock_all(global_win);
	foMPI_Win_free(&local_win);
	foMPI_Win_free(&global_win);
}
