#include "benchmark.h"

using namespace std;

/*
 * latency of RW lock with 4 writers per node
 *
 *  Created on: May 04, 2015
 *      Author: Patrick Schmid
 */
void threshold_rw_latency(MPI_Comm comm, MPI_File filehandle, RWLock rw_lock) {
	HRT_TIMESTAMP_T t[2];
	unsigned long long time_diff;
	
	double *time_acquire, *time_release;
	char line[256];
	
	int iterations = MAX_ITERATIONS;
	
	foMPI_Win win_acquire, win_release;
	int rank, comm_size;
		
	/* Initialization */
	MPI_Comm_rank(comm, &rank);
	MPI_Comm_size(comm, &comm_size);
	foMPI_Win_allocate(iterations * sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &time_acquire, &win_acquire);
	foMPI_Win_allocate(iterations * sizeof(double), sizeof(double), MPI_INFO_NULL, comm, &time_release, &win_release);
	
	/* Warm-up */
	int i;
	if(rank % 16 < 4){
		for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
			rw_lock.writer_acquire();
			rw_lock.writer_release();
		} 
	} else {
		for (i = 0; i < ITERATIONS_PER_WARMUP; i++) {
			rw_lock.reader_acquire();
			rw_lock.reader_release();
		}
	}
	
	/* Start test */
	if(rank % 16 < 4){
		for (i = 0; i < iterations; i++) {
			HRT_GET_TIMESTAMP(t[0]);
			
			rw_lock.writer_acquire();
		
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &time_diff);
			time_acquire[i] = HRT_GET_USEC(time_diff);
			HRT_GET_TIMESTAMP(t[0]);
			
			rw_lock.writer_release();
		
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &time_diff);
			time_release[i] = HRT_GET_USEC(time_diff);
		} 
	} else {
		for (i = 0; i < iterations; i++) {
			HRT_GET_TIMESTAMP(t[0]);
			
			rw_lock.reader_acquire();
		
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &time_diff);
			time_acquire[i] = HRT_GET_USEC(time_diff);
			HRT_GET_TIMESTAMP(t[0]);
			
			rw_lock.reader_release();
		
			HRT_GET_TIMESTAMP(t[1]);
			HRT_GET_ELAPSED_TICKS(t[0], t[1], &time_diff);
			time_release[i] = HRT_GET_USEC(time_diff);
		}
	}

	MPI_Barrier(comm);
	if(rank == 0){
		int j, k;
		for(j = 1; j < comm_size; j++){
			/* Get times for acquiring and releasing from each process */
			double buf_acquire[MAX_ITERATIONS];
			double buf_release[MAX_ITERATIONS];
			foMPI_Win_fence(0, win_acquire);
			foMPI_Get(buf_acquire, iterations, MPI_DOUBLE, j, 0, iterations, MPI_DOUBLE, win_acquire);
			foMPI_Win_fence(0, win_acquire);
			foMPI_Win_fence(0, win_release);
			foMPI_Get(buf_release, iterations, MPI_DOUBLE, j, 0, iterations, MPI_DOUBLE, win_release);
			foMPI_Win_fence(0, win_release);

			/* Add times to other results */
			for(k = 0; k < iterations; k++){
				time_acquire[k] += buf_acquire[k];
				time_release[k] += buf_release[k];
			}
		}
			
		/* Write result to a file */
		double time_acquire_total = 0;
		double time_release_total = 0;
		for (k = 0; k < iterations; k++) {
			/* Divide by number of process who used the lock to get average */
			time_acquire[k] /= comm_size;
			time_release[k] /= comm_size;
			time_acquire_total += time_acquire[k];
			time_release_total += time_release[k];
		}
		double time_acquire_avg = time_acquire_total / iterations / USECONDS_FOR_THROUGHPUT;
		double time_release_avg = time_release_total / iterations / USECONDS_FOR_THROUGHPUT;
		snprintf(&line[0], 256, "%20.8f\t%20.8f\t%20.8f", time_acquire_avg, time_release_avg, time_acquire_avg + time_release_avg);
		MPI_File_write(filehandle, &line[0], strlen(&line[0]), MPI_CHAR, MPI_STATUS_IGNORE);
	} else {
		int j;
		for(j = 1; j < comm_size; j++){
			foMPI_Win_fence(0, win_acquire);
			foMPI_Win_fence(0, win_acquire);
			
			foMPI_Win_fence(0, win_release);
			foMPI_Win_fence(0, win_release);
		}
	}
	
	/* clean-up */
	foMPI_Win_free(&win_acquire);
	foMPI_Win_free(&win_release);
}
